# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the LevelOfPackages metric."""

if __name__ == '__main__':  # pragma: no cover
    # metric instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.design_rules.utils.metric import SCK, Metric


class LevelOfPackages(Metric):
    """Implements the metric interface."""

    def __init__(
        self,
        id='id_0127',
        category='Counters',
        label='Number of nested packages',
        description='Number of nested packages.',
    ):
        super().__init__(
            id=id,
            label=label,
            category=category,
            description=description,
            types=None,
            kinds=[SCK.PACKAGE],
        )

    def on_compute_ex(self, package: suite.Package) -> int:
        """Compute the metric for the input object."""
        result = self._max_level(package)
        self.set_result_metric(result)
        return Metric.OK

    def _max_level(self, package: suite.Package) -> int:
        """Recursively compute the level of packages."""
        level = max(self._max_level(_) for _ in package.packages) if package.packages else 0
        return level + 1


if __name__ == '__main__':  # pragma: no cover
    # metric instantiated outside of a package
    LevelOfPackages()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NumberOfPredefOpsInDiagram metric."""

if __name__ == '__main__':  # pragma: no cover
    # metric instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.design_rules.utils.metric import Metric


class NumberOfPredefOpsInDiagram(Metric):
    """Implements the metric interface."""

    def __init__(
        self,
        id='id_0131',
        category='Counters',
        label='Number of predefined operators in a diagram',
        description='Number of graphical primitive operator instances within a diagram.',
    ):
        super().__init__(
            id=id,
            category=category,
            description=description,
            label=label,
            types=[suite.NetDiagram],
            kinds=None,
        )

    def on_compute(self, diagram: suite.NetDiagram) -> int:
        """Compute the metric for the input object."""
        predef_ops = 0
        for element in diagram.presentation_elements:
            if isinstance(element, suite.EquationGE):
                equation = element.equation
                right = equation.right
                # consider graphical instances of predefined operators or textual equations
                if isinstance(right, suite.ExprCall):
                    if not right.operator or element.kind == 'OBJ_LIT':
                        predef_ops += 1
        self.set_result_metric(predef_ops)
        return Metric.OK


if __name__ == '__main__':  # pragma: no cover
    # metric instantiated outside of a package
    NumberOfPredefOpsInDiagram()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the EqInEqSet rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


from typing import Optional

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class EqInEqSet(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0019',
        label='Equation or branch belong to at least one equation set',
        description=(
            'An equation or a branch shall belong to at least one equation set except '
            'following uses cases:\n'
            "* The element is text: textual diagram, transition's action, textual state, etc.\n"
            '* The element is in a state without diagram (Embedded state)'
        ),
        category='Traceability',
        severity=Rule.ADVISORY,
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            types=[suite.Equation, suite.WhenBranch, suite.IfNode],
            has_parameter=False,
            **kwargs,
        )

    def on_check(self, presentable: suite.Presentable, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        state = self.get_owning_state(presentable)
        if state is not None and not state.diagrams:
            status = Rule.OK
        else:
            pe = presentable.presentation_element
            if not pe or isinstance(pe.diagram, suite.TextDiagram):
                status = Rule.OK
            else:
                if presentable.equation_sets:
                    status = Rule.OK
                else:
                    status = Rule.FAILED
                    kind = 'equation' if isinstance(presentable, suite.Equation) else 'branch'
                    message = 'the {} {} shall belong to at least one equation set'
                    self.set_message(message.format(kind, presentable.get_full_path()))
        return status

    def get_owning_state(self, object_: suite.Object) -> Optional[suite.State]:
        """Return the state owning the item, directly or not, if any."""
        owner = object_.owner
        while not isinstance(owner, suite.Operator):
            if isinstance(owner, suite.State):
                return owner
            owner = owner.owner
        return None


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    EqInEqSet()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the LineCrossing rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.geometry import Line2D, Point, Rect2D
from ansys.scade.design_rules.utils.modelling import get_path
from ansys.scade.design_rules.utils.rule import Rule


class GElement:
    """Abstraction for a graphical element to ease the verifications."""

    def __init__(self, x, y, width, height, name_of_edge, element_type, target):
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.name_of_element = name_of_edge
        self.element_type = element_type
        self.target = target


def line_crosses_element(start_line_x, start_line_y, end_line_x, end_line_y, g_element) -> bool:
    """Return whether a line crosses an element."""
    res1 = False
    if isinstance(g_element.target, suite.Edge):
        line1 = Line2D(
            Point(g_element.x, g_element.y),
            Point(g_element.x + g_element.width, g_element.y + g_element.height),
        )
        line2 = Line2D(Point(start_line_x, start_line_y), Point(end_line_x, end_line_y))
        res1 = line1.is_intersect_line(line2)
    else:
        rect = Rect2D(
            Point(g_element.x, g_element.y),
            Point(g_element.x + g_element.width, g_element.y + g_element.height),
        )
        line = Line2D(Point(start_line_x, start_line_y), Point(end_line_x, end_line_y))
        res1 = line.is_intersection_rect(rect)
    return res1


def is_point_inside_area(point_x, point_y, x, y, width, height) -> bool:
    """Return whether a point is within an area."""
    return (
        (point_x >= x) and (point_y >= y) and (point_x <= (x + width)) and (point_y <= (y + height))
    )


class LineCrossing(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0034',
        category='Readability',
        parameter='lines=yes',
        label='LineCrossing',
        severity=Rule.REQUIRED,
        description=(
            'Overlapping of edges with edges or other elements shall be prevented.\n'
            'Element overlapping is prevented by SCADE editor.\n'
            'Setting the parameter to lines=no does not report edges crossing edges.'
        ),
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.NetDiagram],
            kinds=None,
        )

        # list of already found elements. new elements are checked against this list
        self.g_elements = []

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        d = self.parse_values(parameter)
        if d is None:
            message = f"'{parameter}': parameter syntax error"
        else:
            check_lines = d.get('lines')
            if check_lines is None:
                message = f"'{parameter}': missing 'lines' value"
            else:
                self.check_lines = check_lines.lower() == 'yes'
                return Rule.OK

        self.set_message(message)
        return Rule.ERROR

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        # list of already found elements. new elements are checked against this list
        self.g_elements = []

        # operator for reporting shorter paths
        self.operator = object_.data_def
        while not isinstance(self.operator, suite.Operator):
            self.operator = self.operator.owner

        # first add all box elements to gElements list
        for diagram_object in object_.presentation_elements:
            if (
                isinstance(diagram_object, suite.StateMachineGE)
                or isinstance(diagram_object, suite.StateGE)
                or isinstance(diagram_object, suite.IfBlockGE)
                or isinstance(diagram_object, suite.WhenBlockGE)
            ):
                self.add_element(
                    diagram_object.position[0],
                    diagram_object.position[1],
                    diagram_object.size[0],
                    diagram_object.size[1],
                    diagram_object.presentable.name,
                    None,
                    diagram_object.presentable,
                )
            if isinstance(diagram_object, suite.TransitionGE):
                self.add_element(
                    diagram_object.label_pos[0],
                    diagram_object.label_pos[1],
                    diagram_object.label_size[0],
                    diagram_object.label_size[1],
                    'Transition',
                    None,
                    diagram_object.presentable,
                )
            if isinstance(diagram_object, suite.ActionGE):
                self.add_element(
                    diagram_object.position[0],
                    diagram_object.position[1],
                    diagram_object.size[0],
                    diagram_object.size[1],
                    'Action',
                    None,
                    diagram_object.presentable,
                )
            if isinstance(diagram_object, suite.WhenBranchGE):
                self.add_element(
                    diagram_object.position[0],
                    diagram_object.position[1],
                    diagram_object.label_width,
                    0,
                    'Branch',
                    None,
                    diagram_object.presentable,
                )
            if isinstance(diagram_object, suite.EquationGE):
                equation = diagram_object.equation
                element_type = None
                name = ''
                assert equation is not None  # nosec B101  # addresses linter
                element_type = equation.right
                name = element_type.to_string()
                self.add_element(
                    diagram_object.position[0],
                    diagram_object.position[1],
                    diagram_object.size[0],
                    diagram_object.size[1],
                    name,
                    element_type,
                    diagram_object.presentable,
                )

        # then check all edge elements
        for diagram_object in object_.presentation_elements:
            if isinstance(diagram_object, suite.Edge):
                start_x = 0
                start_y = 0
                first_point = True
                first_line_of_edge = True
                last_line_of_edge = False
                current_line = 0
                number_of_lines = len(diagram_object.points) - 1
                for point in diagram_object.points:
                    if (point[0] == 0) and (point[1] == 0):
                        # automatic wiring, not computed yet
                        message = 'Automatic graphical positions not computed'
                        local_id = diagram_object.left_var.get_oid()
                        self.add_rule_status(
                            diagram_object.src_equation.equation, Rule.ERROR, message, local_id
                        )
                        break
                    if first_point:
                        start_x = point[0]
                        start_y = point[1]
                        first_point = False
                    else:
                        current_line += 1
                        # if two consecutive points are the same, do not test.
                        # line intersection test does not work otherwise
                        if not ((start_x == point[0]) and (start_y == point[1])):
                            if current_line >= number_of_lines:
                                last_line_of_edge = True
                            self._check_line_overlapping(
                                start_x,
                                start_y,
                                point[0] - start_x,
                                point[1] - start_y,
                                diagram_object,
                                first_line_of_edge,
                                last_line_of_edge,
                            )
                            first_line_of_edge = False
                            start_x = point[0]
                            start_y = point[1]

        return Rule.NA

    def add_element(self, start_x, start_y, width, height, name_of_element, element_type, target):
        """
        Add a new element to elements list.

        The name of element is used for lines, for example to not test against same element.
        """
        self.g_elements.append(
            GElement(start_x, start_y, width, height, name_of_element, element_type, target)
        )

    def _check_line_overlapping(
        self,
        start_x,
        start_y,
        width,
        height,
        edge: suite.Edge,
        first_line_of_edge: bool,
        last_line_of_edge: bool,
    ):
        start_line_x = 0
        start_line_y = 0
        end_line_x = 0
        end_line_y = 0
        width_line = 0
        height_line = 0

        if width >= 0:
            start_line_x = start_x
            end_line_x = start_x + width
            width_line = width
        else:
            start_line_x = start_x + width
            end_line_x = start_x
            width_line = -width
        if height >= 0:
            start_line_y = start_y
            end_line_y = start_y + height
            height_line = height
        else:
            start_line_y = start_y + height
            end_line_y = start_y
            height_line = -height

        # get name of edge
        name_of_edge = edge.left_var.name

        # get source and destination element of edge
        source_type = None
        destination_type = None
        not_i_source = edge.src_equation.equation
        assert isinstance(not_i_source, suite.Equation)  # nosec B101  # addresses linter
        source_type = not_i_source.right
        not_i_destination = edge.dst_equation.equation
        if isinstance(not_i_destination, suite.Equation):
            destination_type = not_i_destination.right

        # get eContainer of edge, can be derived from sourceType or destinationType.
        # Not from variable of edge because the variable can be defined in lower levels.
        # First container is always Equation.

        edge_container = source_type.owner.owner
        container_name = ''

        # check whether edges lie within their own container (States, Actions)
        if isinstance(edge_container, suite.State) or isinstance(edge_container, suite.Action):
            if isinstance(edge_container, suite.Action):
                cont_action = edge_container.owner
                if isinstance(cont_action, suite.WhenBranch):
                    container_name = f'WhenBranch ({cont_action.pattern.to_string()}): '
                else:
                    container_name = f'IfAction ({cont_action.if_node.expression.to_string()}): '
            else:
                container_name = edge_container.name
                if container_name != '':
                    container_name += ': '
            # find graphical element within gElements
            for g_element in self.g_elements:
                target = g_element.target
                if not isinstance(target, suite.Edge):
                    # if the containers of edge and target of gElement are equal.
                    # Edge is within a container
                    if edge_container == target:
                        if not (
                            is_point_inside_area(
                                start_line_x,
                                start_line_y,
                                g_element.x,
                                g_element.y,
                                g_element.width,
                                g_element.height,
                            )
                            and is_point_inside_area(
                                end_line_x,
                                end_line_y,
                                g_element.x,
                                g_element.y,
                                g_element.width,
                                g_element.height,
                            )
                        ):
                            eq = edge.src_equation.equation
                            message = f'Edge {name_of_edge} outside boundaries'
                            local_id = f'{edge.left_var.get_oid()}:box'
                            self.add_rule_status(eq, Rule.FAILED, message, local_id)

        for g_element in self.g_elements:
            # is gElement Source of Edge
            is_source = False
            # is gElement Destination of Edge
            is_destination = False
            # is gElement line of the same edge
            is_same_edge = False
            # is gElement in the same container or is sub container of this edge
            is_parent = False

            # check whether line belongs to same edge
            is_same_edge = g_element.name_of_element == name_of_edge

            # get elementType. if elementType != null than a SCADE operator with pins is found.
            # If operator belongs to the edge do not test
            g_element_type = g_element.element_type
            if g_element_type is not None:
                if g_element_type == source_type:
                    is_source = first_line_of_edge
                if g_element_type == destination_type:
                    is_destination = last_line_of_edge

            # check whether gElement is parent of edge */
            # get target (means container) of element
            target = g_element.target
            is_edge = isinstance(target, suite.Edge)
            if not is_edge:  # just Edge is mistaken by Geographics.Edge
                # if the containers of edge and gElement are equal gElement cannot be parent.* /
                if not edge_container == target.owner:
                    edge_container_temp = edge_container
                    # check if target is parent of edge container * /
                    while not isinstance(edge_container_temp, suite.Model):
                        if edge_container_temp == target:
                            is_parent = True
                            break
                        edge_container_temp = edge_container_temp.owner

            # do not check lines against lines of the same edge,
            # source element, destination element or parent elements.
            if (
                not is_same_edge
                and not is_source
                and not is_destination
                and not is_parent
                and (self.check_lines or not is_edge)
            ):
                if line_crosses_element(
                    start_line_x, start_line_y, end_line_x, end_line_y, g_element
                ):
                    eq = edge.src_equation.equation
                    if is_edge:
                        local_path = _get_edge_path(self.operator, g_element.target)
                        local_id = (
                            f'{edge.left_var.get_oid()}:{g_element.target.left_var.get_oid()}'
                        )
                    else:
                        # bugged
                        # local_path = self.operator.get_path(g_element.target)
                        local_path = get_path(self.operator, g_element.target)
                        local_id = f'{edge.left_var.get_oid()}:{g_element.target.get_oid()}'
                    message = f'Edge {name_of_edge} crosses {local_path}'
                    self.add_rule_status(eq, Rule.FAILED, message, local_id)

        self.g_elements.append(
            GElement(start_line_x, start_line_y, width_line, height_line, name_of_edge, None, edge)
        )


def _get_edge_path(operator: suite.Operator, edge: suite.Edge) -> str:
    equation = edge.src_equation.equation
    path = get_path(operator, equation)
    # replace the name of the first left variable with edge's variable
    suffix = equation.lefts[0].name + '='
    path = path.rstrip(suffix) + edge.left_var.name
    return path


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    LineCrossing()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NonLibProjects rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class NonLibProjects(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0059',
        category='Project Structure',
        severity=Rule.REQUIRED,
        parameter='Lib',
        description=(
            "This rule checks if all non-top-level operators, types, and constants of Non-'Lib' "
            'projects are located in other packages nested within the root package.'
        ),
        label=(
            'Projects shall contain only one top-level Operator and its used Types and Constants'
        ),
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Package, suite.NamedType, suite.Constant],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        if model.name.startswith(parameter):
            self.set_message(
                'Provided model is located in a library project. '
                'This rule cannot check library projects!'
            )
            self.model_is_library = True
        else:
            self.model_is_library = False

        return Rule.OK

    def on_check(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if self.model_is_library:
            return Rule.OK

        if (
            isinstance(object, suite.Package)
            and isinstance(object.owner, suite.Model)
            and len(object.operators) > 1
        ):
            self.set_message('More than one top-level operator in project!')
            return Rule.FAILED

        if (
            isinstance(object, suite.Constant)
            and isinstance(object.owner, suite.Package)
            and isinstance(object.owner.owner, suite.Model)
        ):
            not_used_at_top_level = True
            if len(object.clients) == 0:
                not_used_at_top_level = False
            else:  # prevents unnecessary iteration when object.clients is empty
                for client in object.clients:
                    if (
                        isinstance(client.owner, suite.Package)
                        and isinstance(client.owner.owner, suite.Model)
                        and client.owner.owner.owner is None
                    ):
                        not_used_at_top_level = False
                        break

            if not_used_at_top_level:
                self.set_message('Constant defined at top-level package is not used at top-level!')
                return Rule.FAILED

        if (
            isinstance(object, suite.NamedType)
            and isinstance(object.owner, suite.Package)
            and isinstance(object.owner.owner, suite.Model)
        ):
            not_used_at_top_level = True
            if len(object.typed_objects) == 0:
                not_used_at_top_level = False
            else:  # prevents unnecessary iteration when object.typed_objects is empty
                for obj in object.typed_objects:
                    if (
                        isinstance(obj.owner, suite.Package)
                        and isinstance(obj.owner.owner, suite.Model)
                        and obj.owner.owner.owner is None
                    ):
                        not_used_at_top_level = False
                        break

            if not_used_at_top_level:
                self.set_message('Type defined at top-level package is not used at top-level!')
                return Rule.FAILED

        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NonLibProjects()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the SeparateFileNameOperators rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class SeparateFileNameOperators(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0082',
        category='User-Defined Operators',
        parameter='type=dechecked',
        label='All operators should have the option Separate File Name checked or unchecked',
        severity=Rule.ADVISORY,
        description=(
            'This rule checks that all operators have the option '
            'Separate File Name checked (``type=checked``) or '
            'unchecked ``type=dechecked``).'
        ),
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Operator],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        d = self.parse_values(parameter)
        if d is None:
            message = f"'{parameter}': parameter syntax error"
        else:
            type = d.get('type')
            if type is None:
                message = f"'{parameter}': missing 'type' value"
            else:
                self.type = type
                return Rule.OK

        self.set_message(message)
        return Rule.ERROR

    def on_check(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if self.type == 'checked':
            if object.storage_unit is None:
                self.set_message("Option 'Separate File Name' should be checked for this Operator!")
                return Rule.FAILED
        elif object.storage_unit is not None:
            self.set_message("Option 'Separate File Name' should be de-checked for this Operator!")
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    SeparateFileNameOperators()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the TypesTakenFromSpecificProject rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.modelling import is_visible
from ansys.scade.design_rules.utils.rule import SCK, Rule


class TypesTakenFromSpecificPackage(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0101',
        category='Modelling',
        severity=Rule.REQUIRED,
        parameter='packs=CHM',
        label='Types from specific package',
        description=(
            'Interface data types from public operators are taken from specific packages.\n'
            "parameter: list of packages separated by comma: e.g.: 'CHM, etc.'"
        ),
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=None,
            kinds=[SCK.INPUT, SCK.HIDDEN, SCK.OUTPUT, SCK.SENSOR],
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        self.allowed_packages = []
        self.allowed_names = {_.strip() for _ in parameter.split(',')}
        self.allowed_packages = {_ for _ in model.all_packages if _.name in self.allowed_names}
        return Rule.OK

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        # if I/O, verify the owning operator is public
        if isinstance(object_.owner, suite.Operator) and not is_visible(object_.owner):
            return Rule.NA

        type_ = object_.type
        if (
            isinstance(type_, suite.NamedType)
            and not type_.is_predefined()
            and not type_.is_generic()
        ):
            if type_.owner not in self.allowed_packages:
                message = 'Type is not taken from given packages ({}): {}'
                self.set_message(
                    message.format(','.join(self.allowed_names), type_.get_full_path())
                )
                return Rule.FAILED

        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    TypesTakenFromSpecificPackage()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the LevelOfDepthOfStructures rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.apitools.query import get_leaf_type
from ansys.scade.design_rules.utils.rule import Rule


class LevelOfDepthOfStructures(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0031',
        category='Modelling',
        severity=Rule.REQUIRED,
        parameter='4',
        description=(
            "A SCADE structure type shall contain at most 'parameter-value' "
            'level of sub structures.\n'
            'var.level1.level2.level3'
        ),
        label='Level of Depth of Structures',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Structure],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        if not parameter.isdigit():
            self.set_message(
                f'Parameter for rule is not an integer or lower than zero: {parameter}'
            )
            return Rule.ERROR

        return Rule.OK

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        assert isinstance(object_, suite.Structure)  # nosec B101  # addresses linter
        depth = self.get_depth(object_)

        if depth > int(parameter):
            message = f'Level of Depth of structure is too large ({depth} > {parameter})'
            self.set_message(message)
            return Rule.FAILED
        return Rule.OK

    def get_depth(self, structure: suite.Structure) -> int:
        """Return the maximum depth of a structure."""
        max_depth = 0
        for field in structure.elements:
            type_ = get_leaf_type(field.type)
            if isinstance(type_, suite.Structure):
                depth = self.get_depth(type_)
                if depth > max_depth:
                    max_depth = depth

        return max_depth + 1


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    LevelOfDepthOfStructures()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the InitializationOfArrays rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class InitializationOfArrays(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0029',
        category='Modelling',
        severity=Rule.REQUIRED,
        description=(
            'If all elements of an array are identical, the initialization shall be '
            'done like this: value ^size\n'
            "Note: An initialization such as ' ', ' ', ' ', etc. "
            "leads to more memory usage than ' ' ^n."
        ),
        label='Initialization of Arrays',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=False,
            default_param='',
            description=description,
            label=label,
            types=[suite.ExprCall],
            kinds=None,
        )

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        violated = False
        ref_value = None

        # 41 = SC_ECK_BLD_VECTOR
        if object_.predef_opr == 41:
            # assert object_.parameters
            ref_value = self.get_expr_value(object_.parameters[0])
            for parameter in object_.parameters[1:]:
                value = self.get_expr_value(parameter)
                # compare to all other elements
                if ref_value != value:
                    break
            else:
                violated = True

        if violated:
            assert ref_value is not None  # nosec B101  # addresses linter
            container = self.get_closest_annotatable(object_)
            error_msg = f'Use + {ref_value}^{len(object_.parameters)} instead of [{ref_value}, ...]'
            identifier = object_.to_string()
            self.add_rule_status(container, Rule.FAILED, error_msg, identifier)

        return Rule.NA

    def get_expr_value(self, element) -> str:
        """Return the value of an expression."""
        if isinstance(element, suite.ExprId):
            reference = element.reference
            if isinstance(reference, suite.Constant):
                # resolve the constant: allow detecting [0, ZERO]
                return self.get_expr_value(element.reference.value)
        elif isinstance(element, suite.ConstValue):
            # remove _i16 etc.
            return element.value.split('_')[0]

        # default: return the string representation of the expression
        return element.to_string()


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    InitializationOfArrays()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides a command line parser for the rules' parameter."""

from argparse import ArgumentParser, Namespace
import shlex
from typing import Optional


class ParameterParser(ArgumentParser):
    """
    Override the exit and error functions to not exit the application.

    Add a method a parse a parameter string.
    """

    def __init__(self, **kwargs):
        # remove the prefix for usage unless specified
        if 'prog' not in kwargs:
            kwargs = kwargs.copy()
            kwargs['prog'] = ''
        super().__init__(**kwargs)

    def error(self, message: str):
        """
        Override ``error`` to store the message instead of exiting.

        Parameters
        ----------
        message : str
            Error message.
        """
        self.messages.append(message)

    def print_help(self):
        """Override ``print_help`` to store the message instead of raising an exception."""
        self.messages.append(self.format_help())

    def exit(self):
        """Override ``exit`` to not exit the application."""
        pass

    def parse_command(self, command: str) -> Optional[Namespace]:
        """
        Parse a parameter using ArgumentParser.

        Returns None if an error is raised, and the messages are
        accessible in the attribute ``messages``.

        Parameters
        ----------
        command : str
            Input command line.

        Returns
        -------
        Optional[Namespace]
            Result of the parsing or None if an exception is raised.
        """
        # side_effect with self.error
        self.messages = []
        args = shlex.split(command, posix=False)
        try:
            ns = super().parse_args(args)
        except TypeError as e:
            self.messages.append(str(e))
            ns = None
        return ns if not self.messages else None

    @property
    def message(self) -> str:
        """Return a text with all the messages."""
        return '\n'.join(self.messages)

    def format_usage(self) -> str:
        """Return the parser's usage without ``[h]`` and trailing empty line."""
        return super().format_usage().strip('\n').replace(' [-h]', '')



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the LLRNature rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.annotations import AnnotationRule, get_first_note_by_type
from ansys.scade.design_rules.utils.rule import Rule


class LLRNature(AnnotationRule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0036',
        label='CE must have a design annotation',
        description=(
            "The Contributing Elements (CE) shall have an annotation 'DesignElement' "
            "with a property 'Nature'.\n"
            "Parameter: '-t': Name of the note type (e.g.: '-t DesignElement')"
        ),
        category='Traceability',
        severity=Rule.ADVISORY,
        types=None,
        parameter='-t DesignElement',
        **kwargs,
    ):
        if not types:
            types = [suite.EquationSet, suite.TextDiagram, suite.State, suite.Transition]
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            types=types,
            has_parameter=True,
            default_param=parameter,
            **kwargs,
        )

        self.note_type = None

    def on_start(self, model: suite.Model, parameter: str) -> int:
        """Get the rule's parameters."""
        # backward compatibility
        parameter = parameter.replace('note=', '-t ') if parameter else ''
        return super().on_start(model, parameter)

    def on_check(self, annotable: suite.Annotable, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        assert self.note_type is not None  # nosec B101  # addresses linter
        note = get_first_note_by_type(annotable, self.note_type)
        if note:
            status = Rule.OK
        else:
            status = Rule.FAILED
            message = f'the Contributing Element shall have an annotation {self.note_type.name}'
            self.set_message(message)
        return status


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    LLRNature()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the RequirementHasLink rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite
import scade.model.traceability.local_trace_model as local_trace
import scade.model.traceability.traceability as trace

from ansys.scade.design_rules.utils.rule import Rule


class RequirementHasLink(Rule):
    """
    Implements the rule interface.

    This rule uses the 2023R1 requirement API
    """

    def __init__(
        self,
        id='id_0100',
        category='Traceability',
        severity=Rule.REQUIRED,
        parameter='withalmgt=true',
        description=(
            'This rule checks if all requirements in the requirement window are linked.\n'
            "parameter: 'withalmgt=' true or false (e.g.: withalmgt=true)\n"
            '* if withalmgt=true, local links are taken into account.\n'
            '* if withalmgt=false, only pushed information is checked.'
        ),
        label='Requirement has traceability link',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Model],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        d = self.parse_values(parameter)
        if d is None:
            message = f"'{parameter}': parameter syntax error"
        else:
            withalmgt = d.get('withalmgt')
            if not withalmgt:
                message = f"'{parameter}': missing 'withalmgt' value"
            else:
                self.withalmgt = withalmgt == 'true'
                return Rule.OK

        self.set_message(message)
        return Rule.ERROR

    def on_check(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        violated = False

        # check if object (model) is the main project and not a library model
        if object.library:
            return Rule.OK

        path = object.descriptor.model_file_name
        almgr_parser = trace.AlmgrParser()
        try:
            almgr_parser.parse(path)
        except Exception:
            return Rule.OK

        almgt_parser = trace.AlmgtParser()
        almgt_parser.parse(path)

        non_linked_reqs = []
        assert almgr_parser.project is not None  # nosec B101  # addresses linter
        for alm_document in almgr_parser.project.alm_documents:
            for requirement in alm_document.requirements:
                number_of_links = len(requirement.incoming_links)

                # update links_temp with almgt information
                if self.withalmgt and almgt_parser.traceability:
                    for link in almgt_parser.traceability.links:
                        if link.req_id == requirement.identifier:
                            if link.action == local_trace.TraceType.ADD_LINK:
                                number_of_links += 1
                            elif link.action == local_trace.TraceType.UPDATE_LINK:
                                # TODO(Jean): needs to be checked
                                # https://github.com/ansys/scade-design-rules/issues/29
                                number_of_links += 1
                            elif link.action == local_trace.TraceType.REMOVE_LINK:
                                number_of_links -= 1

                if number_of_links <= 0:
                    linked = False
                else:
                    linked = True

                if not linked:
                    non_linked_reqs.append(f'{alm_document.name}.{requirement.name}')
                    violated = True

        if violated:
            self.set_message(f'Not all Requirements have a link: {",".join(non_linked_reqs)}')
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    RequirementHasLink()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the AnnNotesForBasicDataTypesInStructures rule."""

from typing import List

import scade.model.suite as suite

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


from ansys.scade.apitools.query import get_cell_type, get_leaf_type, is_array
from ansys.scade.design_rules.utils.annotations import (
    AnnotationRule,
    get_first_note_by_type,
    is_ann_note_value_defined_and_get_value,
)
from ansys.scade.design_rules.utils.modelling import is_numeric
from ansys.scade.design_rules.utils.rule import Rule


class AnnNotesForBasicDataTypesInStructures(AnnotationRule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0005',
        category='annotations',
        severity=Rule.REQUIRED,
        parameter='-t SDD_TopLevel',
        description=(
            'Each sub-element within structure definitions used for component I/O and TPCs '
            'that includes basic data types must be documented in the notes field with the '
            'following information:\n'
            '* Description:\n* Constraints:\n\n'
            'If not Boolean or Char types:\n* Min_Value:\n* Max_Value:\n* Unit(SI):\n\n'
            'parameter:\n'
            "* '-t': Name of the annotation note type (e.g.: '-t SDD_TopLevel')\n"
        ),
        label='AnnotationNotes for basic data types in structures',
        # ease customization
        std_fields=None,
        numeric_fields=None,
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.CompositeElement],
            kinds=None,
        )
        self.std_fields = std_fields if std_fields else ['Description', 'Constraints']
        self.numeric_fields = (
            numeric_fields if numeric_fields else ['Min_Value', 'Max_Value', 'Unit_SI']
        )

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if not isinstance(object_.owner.owner, suite.NamedType):
            # field of an anonymous structure: ignore
            return Rule.NA

        if is_array(object_.type):
            type_ = get_cell_type(object_.type, True)
        else:
            type_ = object_.type
        type_ = get_leaf_type(type_)
        if isinstance(type_, suite.Structure):
            if not isinstance(type_.owner, suite.NamedType):
                # anonymous structure that can't be evaluated
                self.set_message(f'Anonymous structures not supported: {object_.name}')
                return Rule.FAILED
            else:
                return Rule.NA

        violation_text_missing = self._check_annotation(object_, type_)

        if violation_text_missing:
            self.set_message(
                f'Annotation missing for {object_.name}: {", ".join(violation_text_missing)}'
            )
            return Rule.FAILED
        return Rule.OK

    def _check_annotation(self, object_: suite.CompositeElement, type_: suite.Type) -> List[str]:
        """Check the annotation for the given object."""
        violation_text_missing = []

        ann_note = get_first_note_by_type(object_, self.note_type)
        for field in self.std_fields:
            defined, _ = is_ann_note_value_defined_and_get_value(ann_note, field)
            if not defined:
                violation_text_missing.append(field)
        if is_numeric(type_):
            for field in self.numeric_fields:
                defined, _ = is_ann_note_value_defined_and_get_value(ann_note, field)
                if not defined:
                    violation_text_missing.append(field)

        return violation_text_missing


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    AnnNotesForBasicDataTypesInStructures()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the AnnNotesPresentAndNotEmpty rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.annotations import (
    AnnotationRule,
    ArgumentParser,
    get_first_note_by_type,
    is_ann_note_value_defined_and_get_value,
)
from ansys.scade.design_rules.utils.rule import Rule


class AnnNotesPresentAndNotEmpty(AnnotationRule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0108',
        category='annotations',
        severity=Rule.REQUIRED,
        parameter='-t AnnType_1 -n AT1_Field1 AT1_Field2 AT1_Field3',
        types=None,
        kinds=None,
        description=(
            'Check if an element has a specific annotation note type attached to it. '
            'In addition it is checked that the given field elements are present and not empty.\n\n'
            'parameter:\n'
            "* '-t': Name of the annotation note type (e.g.: '-t AnnType')\n"
            "* '-n': Names of annotation note elements "
            '(e.g.: -n AT1_Field1 AT1_Field2 AT1_Field3)'
        ),
        label='AnnotationNotes present and not empty',
    ):
        if not types:
            types = [suite.Constant, suite.Operator]
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=types,
            kinds=kinds,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        # minimal level of backward compatibility
        parameter = parameter.replace(',notes=', ' -n ').replace(';', ' ')
        status = super().on_start(model, parameter)
        if status == Rule.OK:
            assert self.options is not None  # nosec B101  # addresses linter
            self.names = self.options.names

        return status

    def add_arguments(self, parser: ArgumentParser):
        """Declare arguments in addition to the note type."""
        help = 'Names of the note elements'
        parser.add_argument(
            '-n', '--names', help=help, nargs='+', required=True, metavar='<attribute>'
        )

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        # make sure object_ is annotatable since the selected types
        # can be customized
        if not isinstance(object_, suite.Annotable):
            return Rule.NA

        violation_text = self._check_annotation(object_)

        if violation_text:
            self.set_message(f'Annotation error for {object_.name}: {violation_text}')
            return Rule.FAILED
        return Rule.OK

    def _check_annotation(self, object_: suite.Annotable) -> str:
        """Check the annotation for the given object."""
        violations = []
        assert self.note_type is not None  # nosec B101  # addresses linter
        note = get_first_note_by_type(object_, self.note_type)
        if note is None:
            violations.append(f'Note missing: {self.note_type.name}')
        else:
            for name in self.names:
                defined, _ = is_ann_note_value_defined_and_get_value(note, name)
                if not defined:
                    violations.append(f'Name missing: {name}')

        return ', '.join(violations)


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    AnnNotesPresentAndNotEmpty()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Utilities for intersection checks."""


class Point:
    """Point."""

    def __init__(self, x, y):
        self.x = x
        self.y = y


class Rect2D:
    """Rectangle."""

    def __init__(self, p1: Point, p2: Point):
        self.p1 = p1
        self.p2 = p2


class Line2D:
    """Segment."""

    def __init__(self, p1: Point, p2: Point):
        self.p1 = p1
        self.p2 = p2

    def online(self, line, p: Point) -> bool:
        """Return whether p is on the line."""
        if (
            (p.x <= max(line.p1.x, line.p2.x))
            and (p.x >= min(line.p1.x, line.p2.x))
            and (p.y <= max(line.p1.y, line.p2.y))
            and (p.y >= min(line.p1.y, line.p2.y))
        ):
            return True
        return False

    def direction(self, a: Point, b: Point, c: Point) -> int:
        """Return the direction of an angle."""
        val = (b.y - a.y) * (c.x - b.x) - (b.x - a.x) * (c.y - b.y)
        # colinear
        if val == 0:
            return 0
        # anti-clockwise direction
        elif val < 0:
            return 2
        # clockwise direction
        else:
            return 1

    def is_intersect_line(self, line) -> bool:
        """Return whether the line intersects another line."""
        # four direction for two lines and points of line line
        dir1 = self.direction(self.p1, self.p2, line.p1)
        dir2 = self.direction(self.p1, self.p2, line.p2)
        dir3 = self.direction(line.p1, line.p2, self.p1)
        dir4 = self.direction(line.p1, line.p2, self.p2)

        # lines are intersecting
        if dir1 != dir2 and dir3 != dir4:
            return True
        # when p2 of line2 are on the line1
        if dir1 == 0 and self.online(self, line.p1):
            return True
        # when p1 of line2 are on the line1
        if dir2 == 0 and self.online(self, line.p2):
            return True
        # when p2 of line1 are on the line2
        if dir3 == 0 and self.online(line, self.p1):
            return True
        # when p1 of line1 are on the line2
        if dir4 == 0 and self.online(line, self.p2):
            return True

        return False

    def is_intersection_rect(self, rect: Rect2D) -> bool:
        """Return whether the line intersects a rectangle."""
        rect_point1 = Point(rect.p1.x, rect.p2.y)
        rect_point2 = Point(rect.p2.x, rect.p1.y)
        line1 = Line2D(rect.p1, rect_point1)
        line2 = Line2D(rect.p1, rect_point2)
        line3 = Line2D(rect_point1, rect.p2)
        line4 = Line2D(rect_point2, rect.p2)

        res = (
            self.is_intersect_line(line1)
            or self.is_intersect_line(line2)
            or self.is_intersect_line(line3)
            or self.is_intersect_line(line4)
        )
        if res:
            return True

        return False



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the MaximumCallGraphDepth rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

from typing import List

import scade
import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class MaximumCallGraphDepth(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0039',
        category='Structuring',
        severity=Rule.REQUIRED,
        parameter='depth=7,visibility=Public',
        description=(
            "Maximum depth of the call graph shall not exceed 'parameter'.\n"
            'Check is performed on only public or all operators. '
            'Parameter: depth=:maximum value, visibility=Public,ALL '
            '(e.g.: depth=7,visibility=Public)'
        ),
        label='Maximum call graph depth',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Operator],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        d = self.parse_values(parameter)
        if d is None:
            message = f"'{parameter}': parameter syntax error"
        else:
            depth = d.get('depth')
            visibility = d.get('visibility')
            if not depth:
                message = f"'{parameter}': missing 'depth' value"
            elif not visibility:
                message = f"'{parameter}': missing 'visibility' value"
            else:
                if not depth.isdecimal():
                    message = f"'{depth}' is not a number"
                else:
                    self.depth = int(depth)
                    self.visibility = visibility
                    return Rule.OK

        self.set_message(message)
        # scade is a CPython module defined dynamically
        scade.output(message + '\n')  # type: ignore
        return Rule.ERROR

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if self.visibility == 'Public':
            if object_.visibility != 'Public':
                return Rule.NA

        call_graph_max = self._get_max_call_graph(object_)
        level = len(call_graph_max)
        if level > self.depth:
            self.set_message(
                f'Call graph depth too high ({level} > {self.depth}): {".".join(call_graph_max)}'
            )
            return Rule.FAILED

        return Rule.OK

    def _get_max_call_graph(self, operator: suite.Operator) -> List[str]:
        max_sub_call_graph = []
        for sub_op in operator.called_operators:
            sub_call_graph = self._get_max_call_graph(sub_op)
            if len(sub_call_graph) > len(max_sub_call_graph):
                max_sub_call_graph = sub_call_graph
        return [operator.name] + max_sub_call_graph


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    MaximumCallGraphDepth()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NoOperatorCallsInExpressions rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class NoOperatorCallsInExpressions(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0060',
        category='Modelling',
        severity=Rule.REQUIRED,
        parameter='18',
        description=(
            'No operator calls within textual expressions.\n'
            'This rule does not apply for Constants and Transition/IfNode conditions.\n'
            'Exceptions on main level can be given as parameters.\n'
            "parameter: operator IDs separated by comma: e.g.: '18'"
        ),
        label='No operator calls in expressions',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.ExprCall],
            kinds=None,
        )

    def on_start(self, model, parameter):
        """Get the rule's parameters."""
        if model is None:
            self.set_message('No model found, rule cannot be checked.')
            return Rule.ERROR
        self.excepted_operators = [_.strip() for _ in parameter.split(',')]

        return Rule.OK

    def on_check(self, object_, parameter):
        """Return the evaluation status for the input object."""
        violated = False

        # This rule does not apply for Constants
        container = self.get_closest_annotatable(object_)
        if not isinstance(container, suite.Constant):
            # is expression part of an Equation
            equation = object_.equation
            if equation is not None:
                # get graphical element
                equation_ge = equation.equation_ge
                # check whether object is a textual expression. OBJ_LIT most probably OBJECT_LITERAL
                if equation_ge.kind == 'OBJ_LIT' and not self._is_expression_valid(object_):
                    violated = True

        if violated:
            error_msg = f'Textual expression not in line with given format: {object_.to_string()}'
            identifier = object_.to_string()
            self.add_rule_status(container, Rule.FAILED, error_msg, identifier)

        return Rule.NA

    def _is_expression_valid(self, expression) -> bool:
        return str(expression.predef_opr) in self.excepted_operators


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NoOperatorCallsInExpressions()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Provides an enumeration for finer grain discrimination, mostly used for naming rules."""

from enum import Enum

import scade.model.suite as suite


class SCK(Enum):
    """
    Finer grain discrimination, mostly used for naming rules.

    --> address only named objects
    """

    MODEL = (suite.Model, lambda o: True)
    PACKAGE = (suite.Package, lambda o: not isinstance(o, suite.Model))
    TYPE = (suite.NamedType, lambda o: not (o.is_predefined() or o.is_generic()))
    GENERIC_TYPE = (suite.NamedType, lambda o: o.is_generic())
    FIELD = (suite.CompositeElement, lambda o: True)
    SENSOR = (suite.Sensor, lambda o: True)
    CONSTANT = (suite.Constant, lambda o: isinstance(o.owner, suite.Package))
    ENUM_VALUE = (suite.Constant, lambda o: isinstance(o.owner, suite.Enumeration))
    PARAMETER = (suite.Constant, lambda o: isinstance(o.owner, suite.Operator))
    OPERATOR = (suite.Operator, lambda o: True)
    VARIABLE = (suite.LocalVariable, lambda o: o.is_local() and not o.is_internal())
    INPUT = (suite.LocalVariable, lambda o: o.is_input())
    HIDDEN = (suite.LocalVariable, lambda o: o.is_hidden())
    OUTPUT = (suite.LocalVariable, lambda o: o.is_output())
    SIGNAL = (suite.LocalVariable, lambda o: o.is_signal())
    INTERNAL = (suite.LocalVariable, lambda o: o.is_internal())
    DIAGRAM = (suite.Diagram, lambda o: not isinstance(o, suite.TreeDiagram))
    EQ_SET = (suite.EquationSet, lambda o: True)
    STATE_MACHINE = (suite.StateMachine, lambda o: True)
    STATE = (suite.State, lambda o: True)
    IF_BLOCK = (suite.IfBlock, lambda o: True)
    WHEN_BLOCK = (suite.WhenBlock, lambda o: True)



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the SensorNamesUniqueInPackage rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class SensorNamesUniqueInPackage(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0121',
        category='Naming',
        severity=Rule.REQUIRED,
        description='The name of a Sensor shall not be used for variables in the same package.',
        label='The name of a Sensor shall not be used for variables in the same package.',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=False,
            default_param='',
            description=description,
            label=label,
            types=[suite.LocalVariable],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Cache all sensors."""
        self.sensors = {_.name: _ for _ in model.all_sensors}
        return Rule.OK

    def on_check(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        violated = False

        if object.is_internal():
            # no need to check
            return Rule.NA

        sensor = self.sensors.get(object.name)
        if sensor:
            # raise a violation if the variable is defined in the sensor's package
            # or sub-packages
            owner = object.owner
            while owner:
                if owner == sensor.owner:
                    violated = True
                    break
                owner = owner.owner

        if violated:
            assert sensor is not None  # nosec B101  # addresses linter
            self.set_message(f'Variable name is also used for the sensor: {sensor.get_full_path()}')
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    SensorNamesUniqueInPackage()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NoTerminations rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.apitools.expr import Eck
from ansys.scade.design_rules.utils.rule import Rule


class NoTerminations(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0063',
        category='Modelling',
        severity=Rule.ADVISORY,
        parameter='NOIT',
        description=(
            'Terminations should not be used.\n'
            "parameter: 'ALL': all, 'NOIT: do not report terminations at iterator outputs'"
        ),
        label='No terminations',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Equation],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        if parameter not in ['ALL', 'NOIT']:
            self.set_message('Wrong parameter')
            return Rule.ERROR

        return Rule.OK

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if len(object_.lefts) != 1 or object_.lefts[0].name != '_':
            # not a terminator
            return Rule.NA

        # report the source equation, with the pin number, to allow
        # the equation to be located in the graphics
        local_var_connecting_it_and_termination = object_.right.reference
        prev_eqs = local_var_connecting_it_and_termination.definitions
        if not prev_eqs:
            # semantic error, reporting useless and not easy
            return Rule.NA

        # consider the first source equation
        # (more than one is unlikely but possible with textual diagrams and control blocks)
        violated = False
        prev_eq = prev_eqs[0]
        index_pos_of_var = prev_eq.lefts.index(local_var_connecting_it_and_termination)
        # assert index_pos_of_var == local_var_connecting_it_and_termination.left_range
        right = prev_eq.right
        if parameter == 'NOIT' and isinstance(right, suite.ExprCall):
            # ignore 'exit index' and 'exit condition' outputs
            if right.modifier:
                if Eck(right.modifier.predef_opr) in [
                    Eck.MAPFOLDW,
                    Eck.MAPFOLDWI,
                ]:
                    violated = index_pos_of_var > 1
                elif Eck(right.modifier.predef_opr) in [
                    Eck.MAPW,
                    Eck.MAPWI,
                    Eck.FOLDW,
                    Eck.FOLDWI,
                ]:
                    violated = index_pos_of_var > 0
            else:
                violated = True
        else:
            # no exceptions
            violated = True

        if violated:
            error_msg = f'Termination found at output {index_pos_of_var:.0f}'
            identifier = str(index_pos_of_var)
            self.add_rule_status(prev_eq, Rule.FAILED, error_msg, identifier)

        return Rule.NA


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NoTerminations()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Provides a common base class for metrics, with additional services.

This class stubs the original ``Metric`` class for unit tests.
"""

from typing import List, Optional

import scade.model.suite as suite

try:
    # ignore F401: _register_metric made available for metrics, not used here
    # _register_metric is defined dynamically: ignore linter warning
    from scade.tool.suite import _register_metric  # noqa: F401  # type: ignore

    # _Metric defined hereafter is a stub, and thus can't match Metric
    from scade.tool.suite.metrics import Metric as _Metric  # type: ignore
except ImportError:

    class _Metric:
        """Stubs ``Metric`` for units tests."""

        def __init__(self, id: str, *args, **kwargs):
            self.id = id
            self.message = ''
            self.result = 0

        def set_message(self, message: str) -> None:
            """Store the message."""
            self.message = message

        def set_result_metric(self, result: int) -> None:
            """Stub the ``set_result_metric`` method."""
            self.result = result

        # Possible returned values by the following functions
        OK = 0
        FAILED = 1
        ERROR = 2
        # new 2023 R1, stands for NA
        NO_METRIC = 3


from .sck import SCK


class Metric(_Metric):
    """Base class for all metrics."""

    # set NA to OK until it is implemented
    # new 2023 R1: NA is there but named NO_RULE: let's keep NA
    try:
        NA = _Metric.NO_METRIC
    except BaseException:
        NA = 0

    def __init__(
        self,
        id: str,
        label: str,
        category: str,
        description: str,
        types: Optional[List[type]] = None,
        kinds: Optional[List[SCK]] = None,
        **kwargs,
    ):
        # discrimination of SCADE objects when the class isn't enough
        # compute the list of types when kinds is specified
        if kinds:
            types = [kind.value[0] for kind in kinds]
            self.kinds = kinds
        else:
            self.kinds = None
        # store the types, useful for unit tests
        self.types = types
        super().__init__(id, label, category, types, description=description, **kwargs)
        self.violations = {}

    def on_compute(self, object_: suite.Object) -> int:
        """Compute the metric for the input if it is compatible with the registered kinds."""
        if self.accept_kind(object_):
            return self.on_compute_ex(object_)
        else:
            return Metric.NA

    def on_compute_ex(self, object_: suite.Object) -> int:
        """Compute the metric for the input."""
        # must not be called if on_compute is redefined
        return Metric.ERROR

    def accept_kind(self, object_: suite.Object):
        """Return whether the input is compatible with the registered kinds."""
        if not self.kinds:
            return True

        for kind in self.kinds:
            type, filter = kind.value
            if isinstance(object_, type) and filter(object_):
                return True
        return False



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NameStructureType rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class EqSetHasEqs(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0020',
        label='Equation set content',
        description=(
            'An equation set shall contain only:\n'
            '* Equations\n* If nodes\n* When branches\n* Assertions'
        ),
        category='Traceability',
        severity=Rule.ADVISORY,
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            types=[suite.EquationSet],
            has_parameter=False,
            **kwargs,
        )

    def on_check(self, equation_set: suite.EquationSet, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        paths = [
            presentable.get_full_path()
            for presentable in equation_set.presentables
            if not self.accept(presentable)
        ]
        if paths:
            status = Rule.FAILED
            message = 'equation set {} shall not contain \n\t{}'
            self.set_message(message.format(equation_set.name, '\n\t'.join(paths)))
        else:
            status = Rule.OK
        return status

    def accept(self, presentable: suite.Presentable) -> bool:
        """
        Return whether the element is eligible for belonging to an equation set.

        This method can be overridden in a derived class to consider different elements.
        """
        return (
            isinstance(presentable, suite.Equation)
            or isinstance(presentable, suite.Assertion)
            or isinstance(presentable, suite.IfNode)
            or isinstance(presentable, suite.WhenBranch)
        )


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    EqSetHasEqs()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the ForbiddenOperators rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class ForbiddenOperators(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0023',
        category='Modelling',
        severity=Rule.REQUIRED,
        parameter='45, 28',
        description=(
            'Specific predefined SCADE operators shall not be used: merge (45) and When (28).\n'
            'parameter: comma separated list of operator IDs, for example.: 45,28'
        ),
        label='Forbidden Operators',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.ExprCall],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        self.forbidden_operators = [_.strip() for _ in parameter.split(',')]
        return Rule.OK

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if str(object_.predef_opr) in self.forbidden_operators:
            container = self.get_closest_annotatable(object_)
            error_msg = f'Forbidden operator used: {object_.to_string()}'
            identifier = object_.to_string()
            self.add_rule_status(container, Rule.FAILED, error_msg, identifier)

        return Rule.NA


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    ForbiddenOperators()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the CommentHasSpecificHeadlines rule."""

if __name__ == '__main__':
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class CommentHasSpecificHeadlines(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0009',
        category='annotations',
        severity=Rule.REQUIRED,
        parameter='Purpose',
        label='The comment of an Operator shall have specific headlines.',
        types=None,
        description=(
            'The comment of an object shall have specific headlines defined in parameter.\n'
            'parameter: comma separated list of headlines, for example: Purpose, Algorithm'
        ),
    ):
        if not types:
            types = [suite.Operator]
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=types,
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        self.headlines = [_.strip() for _ in parameter.split(',')]
        return Rule.OK

    def on_check_ex(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        violation_no_text_found = False
        violation_not_all_headlines = False
        found_headlines = []
        messages = []

        number_of_headlines_found = 0
        for comment in object.comments:
            headline_found = False
            textline_found = False
            # is current line a headline
            for headline in self.headlines:
                if comment.startswith(headline):
                    headline_found = True
                    found_headlines.append(headline)
            # is current line a proper text
            if not headline_found:
                if not (comment.isspace() or comment == ''):
                    textline_found = True
            else:
                line = comment.lstrip(found_headlines[-1])
                if line.startswith(':'):
                    line = line.lstrip(':')
                if not (line.isspace() or line == ''):
                    textline_found = True

            if headline_found:
                number_of_headlines_found += 1
                if number_of_headlines_found >= 2:
                    violation_no_text_found = True
                    messages.append(f'No comment for {found_headlines[-2]}')
            if textline_found:
                number_of_headlines_found = 0

        if number_of_headlines_found >= 1:
            violation_no_text_found = True
            messages.append(f'No comment for {found_headlines[-1]}')

        if set(found_headlines) != set(self.headlines):
            violation_not_all_headlines = True
            messages.append('Not all headlines found.')

        if violation_no_text_found or violation_not_all_headlines:
            self.set_message(f'Comment not correct: {",".join(messages)}')
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':
    # rule instantiated outside of a package
    CommentHasSpecificHeadlines()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the LLROnly rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

from typing import List

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class LLROnly(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0037',
        label='Tracing elements must be Contributing Elements',
        description='The model elements tracing requirements shall be Contributing Elements (CE).',
        category='Traceability',
        severity=Rule.ADVISORY,
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            # suite.Traceable does not work, use suite.Object
            # types = [suite.Traceable],
            types=[suite.Object],
            has_parameter=False,
            **kwargs,
        )

    def on_check(self, traceable: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if not isinstance(traceable, suite.Traceable):
            print('not traceable item', str(traceable))
            return Rule.NA

        ids = ', '.join(self.get_requirement_ids(traceable))
        if not self.is_llr(traceable) and ids:
            status = Rule.FAILED
            message = f'the element is not a CE: it shall not trace the requirement(s) {ids}'
            self.set_message(message)
        else:
            status = Rule.OK
        return status

    def is_llr(self, traceable: suite.Traceable) -> bool:
        """
        Return whether the element is a contributing element for traceability.

        This method can be overridden in a derived class to consider different elements.
        """
        return type(traceable) in {
            suite.EquationSet,
            suite.State,
            suite.MainTransition,
            suite.ForkedTransition,
            suite.TextDiagram,
        }

    def get_requirement_ids(self, traceable: suite.Traceable) -> List[str]:  # pragma: no cover
        """
        Return the traced requirement IDs.

        This is a separate function to be redefined when ``Traceable.requirement_ids`` is
        not accessible (unit tests)
        """
        return traceable.requirement_ids


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    LLROnly()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NoAnonymousType rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade
import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import SCK, Rule


class NoAnonymousType(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0122',
        label='No anonymous type in the interface',
        description=(
            'Only named types shall be used in root operator / imported operator '
            'interfaces. Anonymous types (example int^3) shall be avoided.\n'
            'The rule applies to all the root operators, or to the root operators '
            'of the specified configuration.'
        ),
        category='Modelling',
        severity=Rule.ADVISORY,
        parameter='configuration=',
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            kinds=[SCK.INPUT, SCK.HIDDEN, SCK.OUTPUT],
            has_parameter=True,
            default_param=parameter,
            **kwargs,
        )
        self.roots = None

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        # restore the default values
        self.roots = None
        # the parameters are optional
        d = self.parse_values(parameter) if parameter else {}
        if d is not None:
            name = d.get('configuration')
            if name:
                # get the project associated to the model
                project = model.project
                configuration = project.find_configuration(name)
                if configuration:
                    self.roots = project.get_tool_prop_def(
                        'GENERATOR', 'ROOTNODE', [], configuration
                    )
            # no error
            return Rule.OK

        message = f"'{parameter}': parameter syntax error"
        self.set_message(message)
        # scade is a CPython module defined dynamically
        scade.output(message + '\n')  # type: ignore
        return Rule.ERROR

    def on_check_ex(self, variable: suite.LocalVariable, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        # local variable must be an io
        if not self._is_root(variable.operator):
            return Rule.NA
        if not isinstance(variable.type, suite.NamedType):
            variable_type_string = variable.type.to_string() if variable.type else '<null>'
            self.set_message(f'The type {variable_type_string} shall not be anonymous')
            return Rule.FAILED
        # check the sub types
        for type_ in variable.type.used_types:
            # if a type is anonymous, its owner swhall be a named type
            if isinstance(type_, suite.Structure):
                if not isinstance(type_.owner, suite.NamedType):
                    message = 'The type {} shall not contain anonymous structure'
                    self.set_message(message.format(variable.type.to_string()))
                    return Rule.FAILED
            elif isinstance(type_, suite.Table):
                # matrix accepted
                if not isinstance(type_.owner, suite.NamedType) and not isinstance(
                    type_.owner, suite.Table
                ):
                    self.set_message(
                        f'The type {variable.type.to_string()} shall not contain anonymous array'
                    )
                    return Rule.FAILED

        return Rule.OK

    def _is_root(self, operator: suite.Operator) -> bool:
        # no configuration specified, consider any root operator
        # which is not polymorphic nor parameterized by size
        if operator.parameters or operator.typevars:
            return False
        return (
            operator.is_imported()
            or not operator.expr_calls
            # configuration specified: consider its root operators
            and (not self.roots or operator.get_full_path().strip('/') in self.roots)
        )


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NoAnonymousType()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NameShallStartWith rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class NameShallStartWith(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0044',
        category='Naming',
        severity=Rule.REQUIRED,
        parameter='eq_',
        types=None,
        description=(
            "Name shall start with 'parameter'.\nParameter: starting string, for example: 'eq_'"
        ),
        label='Name shall start with',
    ):
        if not types:
            types = [suite.Diagram]
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=types,
            kinds=None,
        )

    def on_check(self, object: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if (isinstance(object, suite.Variable) and object.is_internal()) or isinstance(
            object, suite.TreeDiagram
        ):
            return Rule.OK

        if not object.name.startswith(parameter):
            self.set_message(f'Name does not start with "{parameter}"')
            return Rule.FAILED
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NameShallStartWith()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NoPointerBranch rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import re
from typing import Dict, Optional

import scade.model.suite as suite

from ansys.scade.design_rules.utils.command import ParameterParser
from ansys.scade.design_rules.utils.rule import Rule


class NoPointerBranch(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0091',
        label='Pointer flows must not branch',
        description='A flow, which type is a pointer, shall not be used more than once in a scope.',
        category='Modelling',
        severity=Rule.ADVISORY,
        parameter='-t Ptr.*',
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            types=[suite.LocalVariable],
            has_parameter=True,
            default_param=parameter,
            **kwargs,
        )
        self.types_regexp = ''
        # cache for pointer types
        self.cache_types: Dict[suite.Type, Optional[bool]] = {}

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        # minimal level of backward compatibility
        parameter = parameter.replace('types=', '-t ') if parameter else ''
        parser = ParameterParser(prog='')
        parser.add_argument('-t', '--type', metavar='<type>', help='pointer type', required=True)
        options = parser.parse_command(parameter)
        if not options:
            self.set_message(parser.message)
            return Rule.ERROR

        self.types_regexp = options.type
        return Rule.OK

    def on_check(self, variable: suite.LocalVariable, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if self.is_pointer_type(variable.type):
            # must have one use per scope
            scopes = {_.get_scope() for _ in variable.expr_ids}
            if len(scopes) != len(variable.expr_ids):
                message = f'Illegal branch for {variable.name}'
                self.set_message(message)
                return Rule.FAILED

        return Rule.OK

    def is_pointer_type(self, t: suite.Type) -> bool:
        """Return whether the type is a pointer."""
        is_pointer = self.cache_types.get(t)
        if is_pointer is None:
            if isinstance(t, suite.Table):
                is_pointer = self.is_pointer_type(t.type)
            elif isinstance(t, suite.Structure):
                is_pointer = any(self.is_pointer_type(e.type) for e in t.elements)
            elif isinstance(t, suite.NamedType):
                is_pointer = re.fullmatch(
                    self.types_regexp, t.name
                ) is not None or self.is_pointer_type(t.type)
            else:
                is_pointer = False
        self.cache_types[t] = is_pointer
        return is_pointer


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NoPointerBranch()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the IllegalOperationsOnConstants rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.apitools import print
from ansys.scade.design_rules.utils.naming import substitute_names
from ansys.scade.design_rules.utils.rule import Rule


class IllegalOperationsOnConstants(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0028',
        category='Modelling',
        severity=Rule.REQUIRED,
        parameter='1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 20, 21, 22, 23, 24, '
        '25, 33, 40, 43, 44, 46, 47, 51, 52, 60, 63, 67, 68, 69, 70, 71, 72',
        description=(
            'Operator calls with only constant inputs shall not be used '
            'for the operators in parameter.\n'
            'This rule does not apply for Constants and not for Types.'
        ),
        label='Illegal operations on constants',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.Equation, suite.IfNode, suite.Transition],
            kinds=None,
        )

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        self.check_operators = parameter.split(',')
        try:
            self.check_operators = {int(_) for _ in parameter.split(',') if _.strip()}
            return Rule.OK
        except ValueError:
            pass
        message = f"'{parameter}': syntax error"
        print(message)
        return Rule.ERROR

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        expression = None
        if isinstance(object_, suite.Equation):
            expression = object_.right
        elif isinstance(object_, suite.IfNode):
            expression = object_.expression
        elif isinstance(object_, suite.Transition):
            expression = object_.condition
        else:
            return Rule.OK

        # search for expression calls
        if isinstance(expression, suite.ExprCall):
            # constant sub-expressions
            self.found_elements = []
            # resolution of internal_variables
            self.aliases = {}

            if self._search_through_expr_calls_const(expression):
                # the entire expression is constant
                # assert not self.found_elements
                self.found_elements.append(expression.to_string())

            if self.found_elements:
                # replace internal variables by their constants
                elements = [substitute_names(_, self.aliases) for _ in self.found_elements]
                message = 'Illegal operation on Constant found ({})'.format(', '.join(elements))
                self.set_message(message)
                return Rule.FAILED

        return Rule.OK

    def _search_through_expr_calls_const(self, expr_call) -> bool:
        # check only selected operators
        operator = expr_call.predef_opr

        # constant sub-expressions are recorded (side-effect) if the current expression
        # is not constant, otherwise, will be recorded by the caller
        sub_constant_expressions = []
        # if the operator is not part of the checked ones, check the
        # sub-expressions, but do not consider expr_call to be constant
        all_inputs_are_constants = operator in self.check_operators

        operands = expr_call.parameters
        # go through all operands of the expression call
        for operand in operands:
            # resolve only internal variables, only once
            # ==> named constant flows are ignored
            if isinstance(operand, suite.ExprId):
                producer = operand.reference
                if isinstance(producer, suite.LocalVariable) and producer.is_internal():
                    # must be only one and only one definition
                    # assert len(producer.definitions) == 1
                    operand = producer.definitions[0].right
                    # replace the internal variable with its producer and store the alias
                    self.aliases[producer.name] = f'({operand.to_string()})'

            if isinstance(operand, suite.ExprId):
                if not isinstance(operand.reference, suite.Constant):
                    all_inputs_are_constants = False
            elif isinstance(operand, suite.ExprCall):
                if self._search_through_expr_calls_const(operand):
                    sub_constant_expressions.append(operand.to_string())
                else:
                    all_inputs_are_constants = False

        if not all_inputs_are_constants:
            # store the found constant expressions
            self.found_elements.extend(sub_constant_expressions)
        return all_inputs_are_constants


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    IllegalOperationsOnConstants()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the TextualDiagrams rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class TextualDiagrams(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0083',
        category='Diagrams',
        severity=Rule.ADVISORY,
        label='Textual Diagrams should not be used.',
        types=None,
        description=('This rule checks if Textual Diagrams are used in the project.'),
    ):
        if not types:
            types = [suite.TextDiagram, suite.State, suite.Action]
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=False,
            default_param='',
            description=description,
            label=label,
            types=types,
            kinds=None,
        )

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if isinstance(object_, suite.TextDiagram):
            self.set_message('Project should not contain Textual Diagrams.')
            return Rule.FAILED
        elif type(object_) in [suite.State, suite.Action]:
            pe = object_.presentation_element
            if pe and pe.display == 'EmbeddedTextual':
                self.set_message('Project should not contain Textual Diagrams.')
                return Rule.FAILED
        else:
            # defensive programming: types can be redefined
            return Rule.NA
        return Rule.OK


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    TextualDiagrams()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the NumOfOperatorCallsInConditions rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade
import scade.model.suite as suite

from ansys.scade.design_rules.utils.predef import (
    SC_ECK_AND,
    SC_ECK_EQUAL,
    SC_ECK_GEQUAL,
    SC_ECK_GREAT,
    SC_ECK_LEQUAL,
    SC_ECK_LESS,
    SC_ECK_NEQUAL,
    SC_ECK_NOT,
    SC_ECK_OR,
    SC_ECK_PRJ,
    SC_ECK_SHARP,
    SC_ECK_XOR,
)
from ansys.scade.design_rules.utils.rule import Rule


class NumOfOperatorCallsInConditions(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0075',
        category='Modelling',
        severity=Rule.REQUIRED,
        parameter='calls=4,exc=18',
        description=(
            'Number of logical/comparison operator calls within conditions at '
            'Transitions or IfNodes.\n'
            'Number and Exceptions given as parameters: '
            "'calls=number,exc=op1;op2;etc.'"
        ),
        label='Number of operator calls in conditions',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            types=[suite.IfNode, suite.Transition],
            kinds=None,
        )

    def on_start(self, model, parameter):
        """Get the rule's parameters."""
        self.allowed_pred_ops = [
            SC_ECK_PRJ,
            SC_ECK_AND,
            SC_ECK_OR,
            SC_ECK_XOR,
            SC_ECK_NOT,
            SC_ECK_SHARP,
            SC_ECK_LESS,
            SC_ECK_LEQUAL,
            SC_ECK_GREAT,
            SC_ECK_GEQUAL,
            SC_ECK_EQUAL,
            SC_ECK_NEQUAL,
        ]

        d = self.parse_values(parameter)
        if d is None:
            message = f"'{parameter}': parameter syntax error"
        else:
            calls = d.get('calls')
            exc = d.get('exc', '')
            if not calls:
                message = f"'{parameter}': missing 'calls' value"
            else:
                self.number_of_calls = calls
                self.excepted_operators = {_.strip() for _ in exc.split(';')}
                return Rule.OK

        self.set_message(message)
        # scade is a CPython module defined dynamically
        scade.output(message + '\n')  # type: ignore
        return Rule.ERROR

    def on_check(self, object_, parameter):
        """Return the evaluation status for the input object."""
        violated = False
        self.incorrect = False
        self.levels = 0

        # This rule apply for Transitions and IfNodes
        if isinstance(object_, suite.IfNode):
            expression = object_.expression
        else:
            assert isinstance(object_, suite.Transition)  # nosec B101  # addresses linter
            expression = object_.condition

        self._check_number_of_calls(expression)

        if (self.levels > int(self.number_of_calls)) or self.incorrect:
            violated = True

        if violated:
            if self.incorrect:
                self.set_message(
                    f'Non logical or comparison operators used: {expression.to_string()}'
                )
            else:
                self.set_message(f'Too many operators used: {expression.to_string()}')
            return Rule.FAILED

        return Rule.OK

    def _check_number_of_calls(self, expr: suite.Expression):
        if isinstance(expr, suite.ExprCall):
            if expr.predef_opr not in self.allowed_pred_ops:
                self.incorrect = True
                return
            if self._is_expression_exception(expr):
                num_new_ops = 0
            else:
                num_new_ops = len(expr.parameters) - 1
                # for "not", "-" numNewOps can be 0 although 1 operator is present
                if num_new_ops < 1:
                    num_new_ops = 1
            self.levels += num_new_ops
            for parameter in expr.parameters:
                self._check_number_of_calls(parameter)

    def _is_expression_exception(self, expr_call: suite.ExprCall) -> bool:
        return str(expr_call.predef_opr) in self.excepted_operators


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    NumOfOperatorCallsInConditions()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the ModuloDenominatorNonZeroConstant rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


import scade.model.suite as suite

from ansys.scade.design_rules.utils.modelling import get_value_from_const
from ansys.scade.design_rules.utils.rule import Rule


class ModuloDenominatorNonZeroConstant(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0119',
        category='Modelling',
        severity=Rule.REQUIRED,
        label='Modulo operations shall use a non zero constant as denominator.',
        description='Modulo operations shall use a non zero constant as denominator.',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=False,
            default_param='',
            description=description,
            label=label,
            types=[suite.ExprCall],
            kinds=None,
        )

    def on_check(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        violated = False
        message = ''

        expr_call = object_
        if expr_call.predef_opr_id == 'SC_ECK_MOD':
            denominator_expr = expr_call.parameters[1]
            if isinstance(denominator_expr, suite.ExprId):
                reference = denominator_expr.reference
                if isinstance(reference, suite.LocalVariable):
                    equations = reference.definitions
                    for equation in equations:
                        right = equation.right
                        if isinstance(right, suite.ExprId):
                            denominator = right.reference
                        elif isinstance(right, suite.ConstValue):
                            denominator = right
                elif isinstance(reference, suite.Constant):
                    denominator = reference
            elif isinstance(denominator_expr, suite.ConstValue):
                denominator = denominator_expr

            try:
                const_value = get_value_from_const(denominator)
                if const_value == 0:
                    message = 'Modulo has zero as a denominator.'
                    violated = True
            except Exception:
                violated = True
                message = 'Denominator cannot be evaluated.'

        if violated:
            container = self.get_closest_annotatable(object_)
            identifier = object_.to_string()
            self.add_rule_status(container, Rule.FAILED, message, identifier)

        return Rule.NA


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    ModuloDenominatorNonZeroConstant()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the AnnNoteConnectedDataForPublicInterface rule."""

import scade
import scade.model.suite as suite

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


from ansys.scade.design_rules.utils.annotations import (
    get_first_note_by_type,
    get_note_type,
    is_ann_note_value_defined_and_get_value,
)
from ansys.scade.design_rules.utils.command import ParameterParser
from ansys.scade.design_rules.utils.modelling import is_visible
from ansys.scade.design_rules.utils.rule import SCK, Rule


class AnnNoteConnectedDataForPublicInterface(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0102',
        category='annotations',
        severity=Rule.REQUIRED,
        parameter='inport=RTB_i,outport=RTB_o',
        description=(
            "All interfaces used inside a public operator shall have the 'ConnectedData' "
            'annotation note.\n'
            'IsPrimary is set for outputs. ConnectorName = Input/Output_name + parameter'
        ),
        label='ConnectedData annotation note for interface',
    ):
        super().__init__(
            id=id,
            category=category,
            severity=severity,
            has_parameter=True,
            default_param=parameter,
            description=description,
            label=label,
            kinds=[SCK.INPUT, SCK.OUTPUT],
        )
        # note type as a feature for unit tests
        self.connected_data = 'ConnectedData'

    def on_start(self, model: suite.Model, parameter: str = '') -> int:
        """Get the rule's parameters."""
        # minimal level of backward compatibility
        parameter = (
            parameter.replace('inport=', '-i ').replace(',', ' ').replace('outport=', '-o ')
            if parameter
            else ''
        )
        self.note_type = get_note_type(model, self.connected_data)
        if not self.note_type:
            message = f"'{self.connected_data}': unknown note type"
        else:
            parser = ParameterParser(prog='')
            parser.add_argument(
                '-i', '--inport', help='Suffix for inputs', required=True, dest='in_port'
            )
            parser.add_argument(
                '-o', '--outport', help='Suffix for outputs', required=True, dest='out_port'
            )
            options = parser.parse_command(parameter)
            if not options:
                message = parser.message
            else:
                self.in_port = options.in_port
                self.out_port = options.out_port
                return Rule.OK

        self.set_message(message)
        # scade is a CPython module defined dynamically
        scade.output(message + '\n')  # type: ignore
        return Rule.ERROR

    def on_check_ex(self, object_: suite.Object, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""
        if not is_visible(object_.owner):
            return Rule.NA

        self.violation_text_missing = []

        if object_.is_input():
            self._check_annotation(object_, self.in_port, False)
        else:
            # assert object_.is_output()
            self._check_annotation(object_, self.out_port, True)

        if self.violation_text_missing:
            self.set_message(f'Annotation wrong: {", ".join(self.violation_text_missing)}')
            return Rule.FAILED

        return Rule.OK

    def _check_annotation(self, object_: suite.Object, port: str, primary: bool):
        """Check the annotation for the given object."""
        note = get_first_note_by_type(object_, self.note_type)
        defined, connected_port = is_ann_note_value_defined_and_get_value(note, 'ConnectedPort')
        if not defined:
            self.violation_text_missing.append('ConnectedPort not defined')
        else:
            if connected_port != port:
                self.violation_text_missing.append(f'ConnectedPort not {port}')
        defined, connector_name = is_ann_note_value_defined_and_get_value(note, 'ConnectorName')
        if not defined:
            self.violation_text_missing.append('ConnectedName not defined')
        else:
            # note: if empty values are accepted for port, do not append '_' when empty
            expected_name = object_.name + '_' + port
            if connector_name != expected_name:
                self.violation_text_missing.append(f'ConnectorName not {expected_name}')
        defined, is_primary = is_ann_note_value_defined_and_get_value(note, 'IsPrimary')
        if not defined:
            self.violation_text_missing.append('IsPrimary not defined')
        elif is_primary != primary:
            self.violation_text_missing.append(f'IsPrimary not {primary}')


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    AnnNoteConnectedDataForPublicInterface()



# -*- coding: utf-8 -*-

# Copyright (C) 2024 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Implements the ScopeOfLocals rule."""

if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    from pathlib import Path
    import sys

    sys.path.append(str(Path(__file__).parent.parent.parent.parent.parent.resolve()))


from typing import List

import scade.model.suite as suite

from ansys.scade.design_rules.utils.rule import Rule


class ScopeOfLocals(Rule):
    """Implements the rule interface."""

    def __init__(
        self,
        id='id_0092',
        label='Declaration scope of a local variable',
        description=(
            'The scope of a local variable shall be the least common '
            'scope containing all its references.'
        ),
        category='Modelling',
        severity=Rule.ADVISORY,
        **kwargs,
    ):
        super().__init__(
            id=id,
            label=label,
            description=description,
            category=category,
            severity=severity,
            types=[suite.LocalVariable],
            has_parameter=False,
            **kwargs,
        )

    def on_check(self, var: suite.LocalVariable, parameter: str = '') -> int:
        """Return the evaluation status for the input object."""

        def _get_scope_path(top: suite.DataDef, object_: suite.Object) -> List[suite.Object]:
            scope = object_
            owner = object_
            path = []
            while scope != top:
                while not isinstance(owner, suite.DataDef):
                    owner = owner.owner
                    if isinstance(owner, suite.Action) and isinstance(
                        owner.owner, suite.Transition
                    ):
                        # skip transition's action
                        owner = owner.owner
                    if isinstance(owner, suite.Transition):
                        main = owner.main_transition
                        if main.transition_kind == 'Strong':
                            # the scope of a strong transition is its state's scope
                            assert isinstance(owner.owner, suite.State)  # nosec B101  # addresses linter
                            owner = owner.owner.owner

                scope = owner
                owner = scope.owner
                path.insert(0, scope)
            return path

        if not (var.is_local() and not var.is_internal() or var.is_signal()):
            # the rule does not apply
            return Rule.NA

        # declaration scope
        top = var.owner
        # build the set of scopes for the r/w references
        paths = [_get_scope_path(top, _) for _ in var.expr_ids]
        paths += [_get_scope_path(top, _) for _ in var.definitions]
        if not paths:
            # the variable is unused
            return Rule.OK

        # each path shall start with top
        # assert {_[0] for _ in paths} == {top}
        # get the least common scope containing all the scopes
        # get an arbitrary element as the reference
        ref = paths.pop()
        done = False
        lcs = None
        while ref and not done:
            scope = ref.pop(0)
            for path in paths:
                if not path or path.pop(0) != scope:
                    # found a difference
                    done = True
                    break
            else:
                lcs = scope
        # lcs can't be None since all paths start with top
        assert lcs is not None  # nosec B101  # addresses linter

        if lcs == top:
            # nothing to declare
            status = Rule.OK
        else:
            status = Rule.FAILED
            message = 'the local variable {} shall be declared in its least common scope {}'
            self.set_message(message.format(var.name, lcs.get_full_path()))
        return status


if __name__ == '__main__':  # pragma: no cover
    # rule instantiated outside of a package
    ScopeOfLocals()
